from __future__ import annotations

from copy import deepcopy
from typing import Callable
import itertools
import pandas as pd
import numpy as np

from lamarck.genes import GeneCollection
from lamarck.utils import hash_cols, VectorialOverloadException


class Population:
    """
    Population class.
    """
    data: pd.DataFrame
    blueprint: Blueprint

    def __init__(self, data: pd.DataFrame, blueprint: Blueprint):
        self.blueprint = blueprint
        self.reset_data(data)

    def __len__(self):
        return self.size

    def __add__(self, other):
        return self.merge(other)

    @property
    def size(self) -> int:
        return len(self.data)

    @staticmethod
    def empty(blueprint: Blueprint):
        """
        Create empty Population.
        """
        data = pd.DataFrame(columns=blueprint.genes.names)
        return Population(data, blueprint)

    def reset_data(self, data: pd.DataFrame) -> None:
        self.data = data[self.blueprint.genes.names]

    def hash_index(self) -> pd.DataFrame:
        """
        Returns the Population Dataset (which is a Pandas DataFrame) with hashed index.
        """
        hashed_index = hash_cols(self.data)
        return self.data.set_index(hashed_index)

    def copy(self) -> Population:
        """
        Returns a new Population with the same data as this one.
        """
        new_data = self.data.copy()
        return Population(new_data, self.blueprint.copy())

    def merge(self, pop: Population) -> Population:
        """
        Returns a new Population by combines data from other population.

        Example
        -------
        new_pop = pop1.merge(pop2)

        Alternative
        -----------
        new_pop = pop1 + pop2
        """
        new_data = pd.concat((self.data, pop.data)).reset_index(drop=True)
        return Population(new_data, self.blueprint.copy())

    def unique(self) -> Population:
        """
        Returns a new population without duplicated Creatures.
        """
        new_data = self.data.drop_duplicates()
        return Population(new_data, self.blueprint.copy())


class BlueprintBuilder:
    """
    Genome Blueprint Builder Class.

    Genome Specifications
    ---------------------
        1. Integer
            1.1. Domain: `tuple`

        2. Float
            2.1. Domain: `tuple`

        3. Categorical
            3.1. Domain: `tuple`

        4. Boolean

        5. Array
            5.1. Domain: `tuple`
            5.2. Length: int

        6. Set
            6.1. Domain: `tuple`
            6.2. Length: int
    """
    _blueprint: dict

    def __init__(self):
        self.reset()

    def reset(self) -> None:
        """
        Resets the Blueprint, making it an empty `dict`.
        """
        self._blueprint = {}

    def get_blueprint(self) -> dict:
        """
        Returns a copy of the actual state of the Blueprint.
        """
        blueprint_dict = deepcopy(self._blueprint)
        return Blueprint(blueprint_dict)

    def add_integer_gene(self, name: str, domain: tuple) -> None:
        """
        Add a Integer gene specs for the blueprint.

        Parameters
        ----------
        :name:      Gene name.
        :domain:    Pair of values for `min` and `max` values.

        Example
        -------
        >>> builder = BlueprintBuilder()
        >>> builder.add_integer_gene(name='x', domain=(0, 10))
        >>> blueprint = builder.get_blueprint()
        >>> blueprint.show()
        - x
            |- type: integer
            |- specs
                |- domain: (0, 10)
        """
        genespecs = {
            'type': 'integer',
            'specs': {'domain': domain}
        }
        self._blueprint.update({name: genespecs})

    def add_float_gene(self, name: str, domain: tuple) -> None:
        """
        Add a Float gene specs for the blueprint.

        Parameters
        ----------
        :name:      Gene name.
        :domain:    Pair of values for `min` and `max` values.

        Example
        -------
        >>> builder = BlueprintBuilder()
        >>> builder.add_float_gene(name='x', domain=(0, 10))
        >>> blueprint = builder.get_blueprint()
        >>> blueprint.show()
        - x
            |- type: float
            |- specs
                |- domain: (0, 10)
        """
        genespecs = {
            'type': 'float',
            'specs': {'domain': domain}
        }
        self._blueprint.update({name: genespecs})

    def add_categorical_gene(self, name: str, domain: tuple) -> None:
        """
        Add a Categorical gene specs for the blueprint.

        Parameters
        ----------
        :name:      Gene name.
        :domain:    `tuple` with all categories.

        Example
        -------
        >>> builder = BlueprintBuilder()
        >>> builder.add_categorical_gene(name='letters', domain=('a', 'b', 'c'))
        >>> blueprint = builder.get_blueprint()
        >>> blueprint.show()
        - letters
            |- type: categorical
            |- specs
                |- domain: ('a', 'b', 'c')
        """
        genespecs = {
            'type': 'categorical',
            'specs': {'domain': domain}
        }
        self._blueprint.update({name: genespecs})

    def add_boolean_gene(self, name: str) -> None:
        """
        Add a Boolean gene specs for the blueprint.

        Parameters
        ----------
        :name:  Gene name.

        Example
        -------
        >>> builder = BlueprintBuilder()
        >>> builder.add_boolean_gene(name='match_flag')
        >>> builder.add_boolean_gene(name='is_random')
        >>> blueprint = builder.get_blueprint()
        >>> blueprint.show()
        - match_flag
            |- type: boolean
            |- specs
        - is_random
            |- type: boolean
            |- specs
        """
        genespecs = {
            'type': 'boolean',
            'specs': {}
        }
        self._blueprint.update({name: genespecs})

    def add_array_gene(self, name: str, domain: list | tuple, length: int) -> None:
        """
        Add an Array gene specs for the blueprint.

        An `Array` Gene is a `Vectorial` where its permitted to repeat a value in the same
        vector.

        Parameters
        ----------
        :name:      Gene name.
        :domain:    `tuple` with all possible categories contained in the vector.
        :length:    Length of the vectors.

        Example
        -------
        >>> builder = BlueprintBuilder()
        >>> builder.add_array_gene(name='sequence',
        ...                        domain=('A', 'T', 'C', 'G'),
        ...                        length=1000)
        >>> blueprint = builder.get_blueprint()
        >>> blueprint.show()
        - sequence
            |- type: array
            |- specs
                |- domain: ('A', 'T', 'C', 'G')
                |- length: 1000
        """
        genespecs = {
            'type': 'array',
            'specs': {'domain': domain,
                      'length': length}
        }
        self._blueprint.update({name: genespecs})

    def add_set_gene(self, name: str, domain: list | tuple, length: int) -> None:
        """
        Add an Set gene specs for the blueprint.

        A `Set` Gene is a `Vectorial` where its NOT permitted to repeat a value in the same
        vector.

        Warning: the vector CANNOT have duplicate values, which means its :length: parameter
        CAN'T be greater than the length of its :domain:.

        Parameters
        ----------
        :name:      Gene name.
        :domain:    `tuple` with all possible categories contained in the vector.
        :length:    Length of the vectors (cannot be grater than the length of its :domain:).

        Example
        -------
        >>> builder = BlueprintBuilder()
        >>> builder.add_set_gene(name='cities',
        ...                      domain=('C1', 'C2', 'C3', 'C4', 'C5', 'C6'),
        ...                      length=5)
        >>> blueprint = builder.get_blueprint()
        >>> blueprint.show()
        - cities
            |- type: set
            |- specs
                |- domain: ('C1', 'C2', 'C3', 'C4', 'C5', 'C6')
                |- length: 5
        """
        check_vector_validity(domain, length)
        genespecs = {
            'type': 'set',
            'specs': {'domain': domain,
                      'length': length}
        }
        self._blueprint.update({name: genespecs})


def check_vector_validity(domain, length):
    domain_lenght = len(domain)
    if domain_lenght < length:
        raise VectorialOverloadException(length, domain_lenght)


class Blueprint:
    """
    Blueprint object. Creates the gene generators for Creature Creation.

    Blueprint Dict Example
    ----------------------
    blueprint_dict = {
        'w': {
            'type': 'float',
            'specs': {'domain': (8, 15)}},
        'x': {
            'type': 'float',
            'specs': {'domain': (0, 10)}},
        'y': {
            'type': 'categorical',
            'specs': {'domain': ('A', 'B', 'C')}},
        'z': {
            'type': 'vectorial',
            'specs': {'domain': (0, 1, 2, 3, 4),
                      'replacement': False,
                      'length': 5}},
        'w': {
            'type': 'set',
            'specs': {'domain': (0, 1, 2, 3, 4),
                      'length': 5}},
        'flag': {
            'type': 'boolean'
            'specs': {}}
    }
    blueprint = Blueprint(blueprint_dict)
    """
    class PopDataGenerator:
        """
        Handles deterministic or random Population dataset creation.
        """
        _bp: Blueprint

        def __init__(self, _bp: Blueprint):
            self._bp = _bp

        def deterministic(self, n: int | dict) -> pd.DataFrame:
            """
            Deterministically generates genetic combinations by iterating through
            all of the genes's linearly divided values according to the :n: number
            of values per gene.

            Parameters
            ----------
            :n:     Number of evenly spaced values (subdivisions). The values are
                    selected according to each Gene's Blueprint. If a `dict` is passed,
                    the it is required to declare all genes as Keys and their number of
                    subdivisions as Values.

                    Subdivisions Examples:
                        1. float: min=10, max=90, n=5
                            - values = (10, 30, 50, 70, 90)
                        2. categorical: domain=('A', 'B', 'C', 'D', 'E'), n=3
                            - values = ('A', 'C', 'E')
                        3. array: domain=(1, 2, 3, 4), length=3, n=4
                            - values = ((1, 1, 1),
                                        (2, 1, 1),
                                        (3, 1, 1),
                                        (4, 1, 1))
                        4. set: domain=(1, 2, 3, 4), length=3, n=4
                            - values = ((1, 2, 3),
                                        (2, 1, 3),
                                        (3, 1, 2),
                                        (4, 1, 2))
                    - The Population size will normally be the number n to the power
                    of the number of genes, except if some duplicate genomes are
                    generated (e.g. 4 genes and n=5 will end up generating 5**4=625
                    different combinations of solutions).

            Examples
            --------
            >>> blueprint_dict = {
            ...     'x': {
            ...         'type': 'float',
            ...         'specs': {'domain': (0, 10)}},
            ...     'y': {
            ...         'type': 'categorical',
            ...         'specs': {'domain': ('A', 'B', 'C')}},
            ...     'z': {
            ...         'type': 'array',
            ...         'specs': {'domain': (0, 1, 2, 3, 4),
            ...                   'length': 5}},
            ...     'w': {
            ...         'type': 'set',
            ...         'specs': {'domain': (0, 1, 2, 3, 4),
            ...                   'length': 5}},
            ...     'flag': {
            ...         'type': 'boolean',
            ...         'specs': {}}
            ... }
            >>> blueprint = Blueprint(blueprint_dict)

            - Example 1: n as an Integer (n=3)

            >>> data = blueprint.populate.deterministic(n=3)
            >>> len(data)
            162

            - Example 2: n as a Dictionary
            >>> n_dict = {'x': 5, 'y': 3, 'z': 6, 'w': 2, 'flag': 2}
            >>> data = blueprint.populate.deterministic(n=n_dict)
            >>> len(data)
            360
            """
            if isinstance(n, int):
                n_dict = {gene: n for gene in self._bp._dict}
            else:
                n_dict = n

            def get_unique_linspace(gene):
                genes = self._bp.genes[gene].get_linspace(n_dict[gene])
                return pd.unique(genes)

            ranges = [get_unique_linspace(gene) for gene in self._bp._dict]
            data = itertools.product(*ranges)
            columns = list(self._bp._dict)
            data = pd.DataFrame(data=data, columns=columns)
            pop_data = self._bp.apply_constraints(data).drop_duplicates()
            return Population(pop_data, self._bp)

        def random(self, n: int, seed: int | None = None) -> pd.DataFrame:
            """
            Randomly generates the population data.

            Parameters
            ----------
            :n:     Number of randomly generated solutions. The values will be randomly
                    generated and assigned to each Genome individually.
                    - The number 'n' will normally end up being the Population size,
                    except if some duplicate genomes are generated or the constraints end
                    up trimming a portion of the data.
            :seed:  Random Number Generator control. If set to 'None', the seed will not
                    be enforced (default: None).

            Examples
            --------
            >>> blueprint_dict = {
            ...     'x': {
            ...         'type': 'float',
            ...         'specs': {'domain': (0, 10)}},
            ...     'y': {
            ...         'type': 'categorical',
            ...         'specs': {'domain': ('A', 'B', 'C')}},
            ...     'z': {
            ...         'type': 'array',
            ...         'specs': {'domain': (0, 1, 2, 3, 4),
            ...                   'length': 5}},
            ...     'w': {
            ...         'type': 'set',
            ...         'specs': {'domain': (0, 1, 2, 3, 4),
            ...                   'length': 5}},
            ...     'flag': {
            ...         'type': 'boolean',
            ...         'specs': {}}
            ... }
            >>> blueprint = Blueprint(blueprint_dict)
            >>> data = blueprint.populate.random(n=1000, seed=123)
            >>> len(data)
            1000
            """
            data_dict = {gene: self._bp.genes[gene].get_random(n, seed).tolist()
                         for gene in self._bp._dict}
            data = pd.DataFrame(data_dict)
            pop_data = self._bp.apply_constraints(data).drop_duplicates()
            return Population(pop_data, self._bp)

    _dict: dict
    _constraints: list
    genes = GeneCollection
    populate: PopDataGenerator

    def __init__(self, blueprint_dict: dict):
        self._dict = blueprint_dict
        self.reset_constraints()
        self.genes = GeneCollection(blueprint_dict)
        self.populate = self.PopDataGenerator(self)

    def __repr__(self):
        strlist = []
        for genename, config in self._dict.items():
            genetype = config['type']
            specs = config['specs']
            specstr = ''.join([f'\n        |- {name}: {val}' for name, val in specs.items()])
            genestr = f'- {genename}\n    |- type: {genetype}\n    |- specs{specstr}'
            strlist.append(genestr)
        return '\n'.join(strlist)

    def show(self) -> None:
        """
        Show blueprint specs.
        """
        print(self)

    def copy(self) -> Blueprint:
        """
        Returns copy of this blueprint.
        """
        bp_dict = deepcopy(self._dict)
        return Blueprint(bp_dict)

    def reset_constraints(self) -> None:
        """
        Resets the Constraints, making it an empty `dict`.
        """
        self._constraints = []

    def add_constraint(self, func: Callable[[], bool]) -> None:
        """
        Add a constraint function fir trimming possible input values based on
        user-defined functions that receive a set of the genes as parameters and
        return a boolean value.

        Parameters
        ----------
        :func:  Python function with same parameters as some (or all) of the genes
                names. Function must return a `bool` value.

        Example
        -------
        - Genes x and y are integers and y cannot be bigger than 2*x
        - Code:
        >>> from lamarck import Blueprint
        >>> blueprint_dict = {
        ...     'x': {
        ...         'type': 'integer',
        ...         'specs': {'domain': (0, 10)}},
        ...     'y': {
        ...         'type': 'float',
        ...         'specs': {'domain': (0, 10)}}
        ... }
        >>> blueprint = Blueprint(blueprint_dict)
        >>> def constraint(x, y): return 2*x >= y
        >>> blueprint.add_constraint(constraint)

        * note how the function :constraint(): has parameters that matches some
          (or all) of the specified genes.
        """
        self._constraints.append(func)

    def apply_constraints(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Apply all registered constraint functions for trimming the data.

        Parameters
        ----------
        :data:  DataFrame that has at least all the columns as the parameters
                specified in all declared constraints.
        """
        f = pd.Series(np.ones(len(data), dtype=bool), index=data.index)
        for constraint in self._constraints:
            varnames = constraint.__code__.co_varnames
            nvars = constraint.__code__.co_argcount
            params = list(varnames[0:nvars])
            f = f & data[params].apply(lambda x: constraint(*x), axis=1)
        return data[f]



from __future__ import annotations
from typing import Callable
from dataclasses import dataclass
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

import numpy as np
import pandas as pd
from tqdm import tqdm
from tqdm.contrib.concurrent import thread_map, process_map

from lamarck.rankcalculator import RankCalculator
from lamarck.population import Population
from lamarck.reproduce import Populator
from lamarck.simconfig import SimulationConfig, ParallelMode
from lamarck.utils import objective_ascending_map, get_outputs


class Optimizer:
    """
    The Optimizer is responsible for storing the all the necessary Simulation
    objects, such as:
        - The relevant simulation datasets
        - The user defined process
    Also it has all the simulation control vars and multiple simulation methods.

    Optimizer Structure
    -------------------
    Optimizer
    ├── TempData as temp_data
    │   ├── outputs
    │   └── objectives
    ├── population
    ├── process: user defined function
    ├── Datasets as datasets
    │   ├── population: input data
    │   ├── results: output data
    │   ├── fitness: criteria data
    │   ├── history: historic fitness from multiple generations
    │   └── simulation: concatenation of :population:, :results: and :fitness:
    ├── run()
    ├── run_multithread()
    ├── run_multiprocess()
    ├── Criteria as apply_fitness
    │   ├── single_criteria()
    │   └── MultiCriteria as multi_criteria
    │       ├── ranked()
    │       └── pareto()
    ├── SimulationConfig as config
    └── Simulator as simulate
        ├── single_criteria()
        └── MultiCriteriaSimulations as multi_criteria
            ├── ranked()
            └── pareto()
    """
    @dataclass
    class TempData:
        outputs: list[str] | str | None = None
        objectives: list[str] | str | None = None

    class Datasets:
        """
        Pertinent datasets for the simulation control.

        Datasets
        --------
        - population:   Gene data for all creatures in the population ("input" data);
        - results:      Results of the creatures according to each of the user-defined
                        function outputs ("output" data);
        - fitness:      Population Rankings defined by the defined criteria
                        ("performance" data);
        - simulation:   Concatenation of the main datasets (population-results-fitness);
        - history:      Simulation data from previous generations.
        """
        _opt: Optimizer
        _outputs: tuple
        population: pd.DataFrame
        results: pd.DataFrame
        fitness: pd.DataFrame
        history: pd.DataFrame

        def __init__(self, opt: Optimizer):
            self._opt = opt
            self._outputs = get_outputs(opt.process)
            self._build_initial_data()

        @property
        def simulation(self):
            datasets = (self.population, self.results, self.fitness)
            return pd.concat(datasets, axis=1)

        def _build_initial_data(self) -> None:
            """
            Create the initial Simulation Datasets.

            Datasets
            --------
            :population:    create dataset by copying the data from the Optimizer's :population:
                            attribute and changing the index to the genes's hashs.
            :results:       create dataset with the same hashed index and with empty
                            columns named after the :_outputs: attribute values.
            :fitness:       create empty dataset with the hashed index.
            :history:       create empty dataset with the hashed index.
            """
            self.update_data()
            self.reset_history_data(pd.DataFrame())

        def update_data(self):
            df = self._opt.population.hash_index()
            self.population = df
            self.results = pd.DataFrame(columns=self._outputs, index=df.index)
            self.fitness = pd.DataFrame(index=df.index)

        def assign_fitness(self, fitness_data: pd.DataFrame) -> None:
            self.fitness = fitness_data.copy()

        def reset_history_data(self, history: pd.DataFrame, generation: int = 0) -> None:
            self.history = history.assign(generation=generation)

        def add_history_data(self, history: pd.DataFrame, generation: int = 0) -> None:
            self.history = pd.concat((self.history, history.assign(generation=generation)))

        def get_best_criature(self,
                              outputs: list | str | None = None,
                              objectives: list | str | None = None) -> pd.Series:
            """
            Get the best creature from the simulation results by selecting the columns to sort
            and the objectives ('min' or 'max') to determine how to sort them.

            If arguments are `None`, the temp 'outputs' and 'objectives' from this Optimizer
            object will be used.
            """
            if outputs is None:
                outputs = self._opt.temp_data.outputs
            if isinstance(outputs, str):
                outputs = [outputs]
            if objectives is None:
                objectives = self._opt.temp_data.objectives
            if isinstance(objectives, str):
                objectives = [objectives]
            ascending = [objective_ascending_map[objective] for objective in objectives]
            return self.simulation.sort_values(outputs, ascending=ascending).iloc[0]

    class Criteria:
        """
        Apply a Fitness Criteria on selected output(s), defining the objective
        (maximize/minimize).

        Available Criteria Methods
        --------------------------
        Single Objective            Select an output to maximize or minimize;
        Multi Objective - Ranked    Select multiple outputs and maximize/minimize them in
                                    order of priority;
        Multi Objective - Pareto    Select multiple outputs and find their Pareto Fronts
                                    by attributing a set of maximize/minimize objectives.
        """
        _opt: Optimizer

        def __init__(self, opt: Optimizer):
            self._opt = opt

        def single_criteria(self, output: str, objective: str) -> None:
            """
            Select one :output: to optimize according to a defined :objective:.

            Parameters
            ----------
            :output:       Output Variable.
            :objective:    Objective.

            Available Objectives
            --------------------
            'min'   Minimize the output (best genes are the ones that return lower values);
            'max'   Maximize the output (best genes are the ones that return greater values)

            Updates
            -------
            self.datasets.results DataFrame
            """
            self._opt._update_temp_data(output, objective)
            self._opt.rank_calculator.update(self._opt.datasets.results, output)
            rank = self._opt.rank_calculator.single(objective)
            concat_data = (
                self._opt.datasets.results[output].rename('Criteria'),
                rank
            )
            fitness_data = pd.concat(concat_data, axis=1)
            self._opt.datasets.assign_fitness(fitness_data)

        def multi_criteria_ranked(self, outputs: list[str], objectives: list[str]) -> None:
            """
            Select multiple :outputs: to optimize according to a set of defined :objectives:
            in order of priority.

            Parameters
            ----------
            :outputs:       Output Variables.
            :objectives:    Objectives.

            Available Objectives
            --------------------
            'min'   Minimize the output (best genes are the ones that return lower values);
            'max'   Maximize the output (best genes are the ones that return greater values)

            Example
            -------
            opt.apply_fitness.multi_criteria.ranked(outputs=['x', 'y']
                                                    objectives=['max', 'min'])
            ...will rank according to the biggest 'x' values, then to the lowest 'y' values.

            Updates
            -------
            self.datasets.results DataFrame
            """
            self._opt._update_temp_data(outputs, objectives)
            self._opt.rank_calculator.update(self._opt.datasets.results, outputs)
            rank = self._opt.rank_calculator.ranked(objectives)
            concat_data = (
                [self._opt.datasets.results[col].rename(f'Criteria{i+1}')
                    for i, col in enumerate(outputs)]
                + [rank]
            )
            fitness_data = pd.concat(concat_data, axis=1)
            self._opt.datasets.assign_fitness(fitness_data)

        def multi_criteria_pareto(self, outputs: list[str], objectives: list[str]) -> None:
            """
            Select multiple :outputs: to optimize according to a set of defined :objectives:
            that will define the Pareto Fronts. The Ranking method will consider both the
            `front` and the `crowd` metrics of all genes.

            Parameters
            ----------
            :outputs:       Output Variables.
            :objectives:    Objectives.

            Available Objectives
            --------------------
            'min'   Minimize the output (best genes are the ones that return lower values);
            'max'   Maximize the output (best genes are the ones that return greater values)

            Example
            -------
            opt.apply_fitness.multi_criteria.pareto(outputs=['x', 'y']
                                                    objectives=['max', 'min'])
            ...will rank according to the pareto fronts (ascending) and then the `crowd`
            values (descending).

            Updates
            -------
            self.datasets.results DataFrame
            """
            self._opt._update_temp_data(outputs, objectives)
            self._opt.rank_calculator.update(self._opt.datasets.results, outputs)
            fronts = self._opt.rank_calculator.pareto_fronts(objectives)
            crowd = self._opt.rank_calculator.pareto_crowds(fronts)
            rank = self._opt.rank_calculator.pareto(objectives)
            concat_data = (fronts, crowd, rank)
            fitness_data = pd.concat(concat_data, axis=1)
            self._opt.datasets.assign_fitness(fitness_data)

    class Simulator:
        """
        Run genetic simulations by selecting a Fitness Criteria on selected output(s),
        defining the objectives (maximize/minimize).

        The simulation will run multiple generations and perform a selection criteria,
        then reproductions and mutations, all according to the `config` settings.

        Available Criteria Methods
        --------------------------
        Single Objective            Select an output to maximize or minimize;
        Multi Objective - Ranked    Select multiple outputs and maximize/minimize them in
                                    order of priority;
        Multi Objective - Pareto    Select multiple outputs and find their Pareto Fronts
                                    by attributing a set of maximize/minimize objectives.
        """
        class MultiCriteriaSimulations:
            """
            Multi-objective simulations
            ---------------------------
            - Ranked
            - Pareto
            """
            _opt: Optimizer

            def __init__(self, opt):
                self._opt = opt

            @staticmethod
            def _stall_criteria(
                    outputs: list[str],
                    objectives: list[str]) -> Callable[[int, pd.DataFrame, pd.DataFrame], int]:
                def check_stall(n_stall: int,
                                new_pop: pd.DataFrame,
                                old_pop: pd.DataFrame) -> int:
                    ascendings = [objective_ascending_map[objective] for objective in objectives]
                    new_best = new_pop.sort_values(outputs, ascending=ascendings).iloc[0]
                    old_best = old_pop.sort_values(outputs, ascending=ascendings).iloc[0]
                    stall_condition = True
                    for output, objective in zip(outputs, objectives):
                        ascending = objective_ascending_map[objective]
                        if ascending:
                            stall_condition &= new_best[output] >= old_best[output]
                        else:
                            stall_condition &= new_best[output] <= old_best[output]
                    return n_stall+1 if stall_condition else 0
                return check_stall

            def ranked(self,
                       outputs: list[str],
                       objectives: list[str],
                       quiet: bool = False,
                       seed: int | None = None) -> None:
                """
                Run optimization with multiple objectives, selecting the best of them with the
                Ranked Criteria.

                Parameters
                ----------
                :outputs:       The process outputs that will be optimized
                :objectives:    The output objectives ('min' for `minimizing` or 'max' for
                                `maximizing`)
                :quiet:         If `False`, a bar will show the simulation progress (default:
                                `False`).
                :seed:          Random Number Generator control (default: None).
                """
                np.random.seed(seed)
                self._opt._update_temp_data(outputs, objectives)

                def apply_fitness(opt: Optimizer) -> None:
                    return opt.apply_fitness.multi_criteria_ranked(outputs=outputs,
                                                                   objectives=objectives)

                check_stall = self._stall_criteria(outputs, objectives)
                optimize(self._opt, apply_fitness, check_stall, quiet)

            def pareto(self,
                       outputs: list[str],
                       objectives: list[str],
                       quiet: bool = False,
                       seed: int | None = None) -> None:
                """
                Run optimization with multiple objectives, selecting the best of them with the
                Pareto Criteria.

                Parameters
                ----------
                :outputs:       The process outputs that will be optimized
                :objectives:    The output objectives ('min' for `minimizing` or 'max' for
                                `maximizing`)
                :quiet:         If `False`, a bar will show the simulation progress (default:
                                `False`).
                :seed:          Random Number Generator control (default: None).
                """
                np.random.seed(seed)
                self._opt._update_temp_data(outputs, objectives)

                def apply_fitness(opt: Optimizer) -> None:
                    return opt.apply_fitness.multi_criteria_pareto(outputs=outputs,
                                                                   objectives=objectives)

                check_stall = self._stall_criteria(outputs, objectives)
                optimize(self._opt, apply_fitness, check_stall, quiet)

        _opt: Optimizer
        multi_criteria: MultiCriteriaSimulations

        def __init__(self, opt: Optimizer):
            self._opt = opt
            self.multi_criteria = self.MultiCriteriaSimulations(opt)

        def single_criteria(self,
                            output: str,
                            objective: str,
                            quiet: bool = False,
                            seed: int | None = None) -> None:
            """
            Run optimization with a single objective.

            Parameters
            ----------
            :output:    The process output that will be optimized
            :objective: The output objective ('min' for `minimizing` or 'max' for `maximizing`)
            :quiet:     If `False`, a bar will show the simulation progress (default: `False`).
            :seed:      Random Number Generator control (default: None).
            """
            np.random.seed(seed)
            self._opt._update_temp_data(output, objective)

            def apply_fitness(opt: Optimizer) -> None:
                return opt.apply_fitness.single_criteria(output=output, objective=objective)

            def check_stall(n_stall: int, new_pop: pd.DataFrame, old_pop: pd.DataFrame) -> int:
                ascending = objective_ascending_map[objective]
                new_best = new_pop[output].sort_values(ascending=ascending).iloc[0]
                old_best = old_pop[output].sort_values(ascending=ascending).iloc[0]
                if ascending:
                    stall_condition = new_best >= old_best
                else:
                    stall_condition = new_best <= old_best
                return n_stall+1 if stall_condition else 0

            optimize(self._opt, apply_fitness, check_stall, quiet)

    _original_population_data: pd.DataFrame
    population: Population
    process: Callable[[], dict]
    datasets: Datasets
    apply_fitness: Criteria
    rank_calculator: RankCalculator
    config: SimulationConfig
    simulate: Simulator

    def __init__(self, population: Population, process: Callable[[], dict]):
        self.temp_data = self.TempData()
        self.process = process
        self.set_population(population)
        self.apply_fitness = self.Criteria(self)
        self.rank_calculator = RankCalculator()
        self.config = SimulationConfig()
        self.simulate = self.Simulator(self)

    def _update_temp_data(self, outputs: list[str] | str, objectives: list[str] | str):
        self.temp_data.outputs = outputs
        self.temp_data.objectives = objectives

    def clear_datasets(self):
        self.datasets = self.Datasets(self)

    def set_population(self, population: Population) -> None:
        self._original_population_data = population.data.copy()
        self.population = population
        self.clear_datasets()

    def reset(self):
        self.reset_population_data(self._original_population_data.copy())

    def reset_population_data(self, data: pd.DataFrame) -> None:
        self.population.reset_data(data)
        self.clear_datasets()

    def update_population(self, population: Population) -> None:
        self.population = population
        self.datasets.update_data()

    def run(self, quiet: bool = False) -> None:
        """
        Run the process for all genes in the Population.

        Parameters
        ----------
        :quiet: If `False`, a bar will show the simulation progress (default: `False`).

        Updates
        -------
        self.datasets.results DataFrame
        """
        def run_generator_quiet():
            for index, row in self.datasets.population.iterrows():
                yield pd.DataFrame(self.process(**row), index=pd.Series([index], name='id'))

        def run_generator():
            bar = tqdm(total=self.population.size, position=0, desc='Simulating Population')
            for index, row in self.datasets.population.iterrows():
                yield pd.DataFrame(self.process(**row), index=pd.Series([index], name='id'))
                bar.update()
            bar.close()

        if quiet:
            sim_generator = run_generator_quiet()
        else:
            sim_generator = run_generator()
        self.datasets.results = pd.concat(sim_generator)

    def run_multithread(self,
                        max_workers: int | None = None,
                        quiet: bool = False) -> None:
        """
        Run the process for all genes in the Population with Multi Threading.

        Parameters
        ----------
        :max_workers:   Number of concurrent workers. If `None`, the algorithm will get the
                        maximum possible (default: `None`).
        :quiet:         If `False`, a bar will show the simulation progress (default: `False`).

        Updates
        -------
        self.datasets.results DataFrame
        """
        def sim(x):
            index, row = x
            return pd.DataFrame(self.process(**row), index=pd.Series([index], name='id'))

        iterable = self.datasets.population.iterrows()
        if quiet:
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                creature_result_list = executor.map(sim, iterable)
        else:
            population_size = self.population.size
            creature_result_list = thread_map(sim,
                                              iterable,
                                              total=population_size,
                                              position=1,
                                              desc='Simulating Population',
                                              max_workers=max_workers)
        self.datasets.results = pd.concat(creature_result_list)

    def run_multiprocess(self,
                         max_workers: int | None = None,
                         quiet: bool = False) -> None:
        """
        Run the process for all genes in the Population with Multi Processing.

        Parameters
        ----------
        :max_workers:   Number of concurrent workers. If `None`, the algorithm will get the
                        maximum possible (default: `None`).
        :quiet:         If `False`, a bar will show the simulation progress (default: `False`).

        Updates
        -------
        self.datasets.results DataFrame
        """
        iterable = self.datasets.population.iterrows()
        if quiet:
            with ProcessPoolExecutor(max_workers=max_workers) as executor:
                creature_result_list = executor.map(self._sim_multiprocess, iterable)
        else:
            population_size = self.population.size
            creature_result_list = process_map(self._sim_multiprocess,
                                               iterable,
                                               total=population_size,
                                               position=1,
                                               desc='Simulating Population',
                                               max_workers=max_workers)
        self.datasets.results = pd.concat(creature_result_list)

    def _sim_multiprocess(self, x):
        index, row = x
        return pd.DataFrame(self.process(**row), index=pd.Series([index], name='id'))


def select_fittest(ranked_pop_data: pd.DataFrame,
                   p: float,
                   rank_col: str = 'Rank') -> pd.DataFrame:
    """
    Select the fraction :p: of the fittest creatures from a ranked population data based on a
    Rank column.
    """
    n_selection = int(round(len(ranked_pop_data) * p))
    return ranked_pop_data.sort_values(rank_col)[0:n_selection]


def select_weakest(ranked_pop_data: pd.DataFrame,
                   p: float,
                   rank_col: str = 'Rank') -> pd.DataFrame:
    """
    Select the fraction :p: of the weakest creatures from a ranked population data based on a
    Rank column.
    """
    n_selection = int(round(len(ranked_pop_data) * p))
    return ranked_pop_data.sort_values(rank_col)[-n_selection:]


def select(opt: Optimizer):
    selected_data = select_fittest(ranked_pop_data=opt.datasets.simulation,
                                   p=opt.config.p_selection)
    weakest_data = select_weakest(ranked_pop_data=opt.datasets.simulation,
                                  p=1-opt.config.p_selection)
    n_weaker = int(len(opt.datasets.simulation) * opt.config.p_selection_weak)
    weaker_data = weakest_data.sample(n_weaker)
    return pd.concat((selected_data, weaker_data))


def get_minimum_population(config: Optimizer.SimulationConfig) -> int:
    """
    Get the minimum amount of creatures in a Population required to properly follow the
    simulation's specifications.
    """
    return int(np.ceil(config.n_parents + config.n_dispute - 1) / config.p_selection)


def run_sim(opt: Optimizer, quiet: bool) -> Optimizer:
    """
    Creates a copy of the Optimizer and use it to run a simulation for later criteria
    selection.
    """
    if opt.config.parallel_processing == ParallelMode.OFF:
        opt.run(quiet=quiet)
    elif opt.config.parallel_processing == ParallelMode.MULTITHREAD:
        opt.run_multithread(max_workers=opt.config.max_workers, quiet=quiet)
    elif opt.config.parallel_processing == ParallelMode.MULTIPROCESS:
        opt.run_multiprocess(max_workers=opt.config.max_workers, quiet=quiet)
    else:
        raise Exception(f"Invalid config {opt.config.parallel_processing}.")
    return opt


def generate_offspring_sexually(opt: Optimizer,
                                n_offspring: int,
                                parent_fitness_data: pd.DataFrame):
    reproduce = Populator(opt.population.blueprint)
    if n_offspring > 0:
        offspring_sexual = reproduce.sexual(
            ranked_pop_data=parent_fitness_data,
            n_offspring=n_offspring,
            n_dispute=opt.config.n_dispute,
            n_parents=opt.config.n_parents,
            children_per_relation=opt.config.children_per_relation)
        new_pop_sexual = offspring_sexual
    else:
        new_pop_sexual = Population.empty(opt.population.blueprint)
    return new_pop_sexual


def generate_offspring_asexually(opt: Optimizer,
                                 n_offspring: int,
                                 parent_fitness_data: pd.DataFrame):
    reproduce = Populator(opt.population.blueprint)
    if n_offspring > 0:
        offspring_mutation = reproduce.asexual(
            ranked_pop_data=parent_fitness_data,
            n_offspring=n_offspring,
            n_mutated_genes=opt.config.max_mutated_genes,
            children_per_creature=opt.config.children_per_mutation)
        new_pop_asexual = offspring_mutation
    else:
        new_pop_asexual = Population.empty(opt.population.blueprint)
    return new_pop_asexual


def make_new_generation(n: int,
                        parent_fitness_data: pd.DataFrame,
                        p_mutation: float,
                        p_random: float,
                        opt: Optimizer) -> Population:
    p_selection = opt.config.p_selection - opt.config.p_selection_weak
    # sexual
    p_sexual = 1 - p_selection - p_mutation - p_random
    n_offspring_sexual = int(np.ceil(p_sexual * n))
    new_pop_sexual = generate_offspring_sexually(opt, n_offspring_sexual, parent_fitness_data)
    # asexual
    n_offspring_asexual = int(np.ceil(p_mutation * n))
    new_pop_asexual = generate_offspring_asexually(opt, n_offspring_asexual, parent_fitness_data)
    # random
    n_offspring_random = int(np.ceil(p_random * n))
    new_pop_random = opt.population.blueprint.populate.random(n_offspring_random)
    return new_pop_sexual + new_pop_asexual + new_pop_random


def repopulate(opt: Optimizer, selected_data: pd.DataFrame, target_size: int) -> Population:
    """
    Create new generation.
    """
    fittest_pop = Population(selected_data, opt.population.blueprint)
    offspring_pop = make_new_generation(n=target_size,
                                        parent_fitness_data=selected_data,
                                        p_mutation=opt.config.p_mutation,
                                        p_random=opt.config.p_new_random,
                                        opt=opt)
    new_pop = (fittest_pop + offspring_pop).unique()
    if (new_pop.size < target_size) and opt.config.randomize_to_fill_pop:
        pop_short = target_size - new_pop.size
        complemental_pop = make_new_generation(n=pop_short,
                                               parent_fitness_data=selected_data,
                                               p_mutation=0,
                                               p_random=1,
                                               opt=opt)
        new_pop = (new_pop + complemental_pop).unique()
    return new_pop


def get_descriptor(config):
    if config.peek_champion_variables is None:
        def descriptor(opt, generation, n_stall):
            return f'Generation {generation} of {config.max_generations} '\
                   f'(stall: {n_stall} of {config.max_stall})'
    else:
        def descriptor(opt, generation, n_stall):
            best_creature = opt.datasets.get_best_criature()
            peek_info = best_creature[config.peek_champion_variables].iteritems()
            peek = ' | '.join([f'{feat}: {val}' for feat, val in peek_info])
            return f'Generation {generation} of {config.max_generations} '\
                   f'(stall: {n_stall} of {config.max_stall}) '\
                   f'peek: ({peek})'
    return descriptor


def optimize(opt: Optimizer,
             apply_fitness: Callable[[], None],
             check_stall: Callable[[int, pd.DataFrame, pd.DataFrame], int],
             quiet: bool) -> None:
    """
    Run simulation multiple times until the best Creatures are found.
    """
    opt.reset()
    n_stall = 0
    config = opt.config
    minimum_pop = get_minimum_population(config)
    old_sim_data = opt.datasets.simulation.copy()
    target_size = opt.population.size
    descriptor = get_descriptor(config)
    pbar = tqdm(total=config.max_generations, position=0)
    for generation in range(1, config.max_generations + 1):
        # simulate
        simulated_opt = run_sim(opt, quiet)
        apply_fitness(opt)
        # add history to optimizer
        opt.datasets.add_history_data(opt.datasets.simulation, generation)
        # update bar description
        descr = descriptor(opt, generation, n_stall)
        pbar.set_description(descr)
        pbar.update()
        # check stall
        new_sim_data = opt.datasets.simulation.copy()
        n_stall = check_stall(n_stall, new_sim_data, old_sim_data)
        if (n_stall > config.max_stall) or (generation == config.max_generations):
            break
        else:
            # select
            selected_data = select(simulated_opt)
            # repopulate
            new_pop = repopulate(simulated_opt, selected_data, target_size)
            if new_pop.size < minimum_pop:
                break
            # update opt's population
            opt.update_population(new_pop)
            old_sim_data = new_sim_data
    pbar.close()



from __future__ import annotations
from typing import Callable

import pandas as pd
import matplotlib

from lamarck.optimizer import Optimizer


class DataExplorer:
    """
    Base Data Explorer class.
    """
    opt: Optimizer

    def __init__(self, opt: Optimizer):
        self.opt = opt

    @property
    def main_data(self) -> pd.DataFrame:
        return self.opt.datasets.population

    def variable_pair(self, x: str, y: str, **kw) -> matplotlib.axes.Axes:
        """
        Scatter plot of two variables.

        Parameters
        ----------
        :x:     `str` for the X axis column.
        :y:     `str` for the Y axis column.
        Key-Word Arguments - Pandas DataFrame Scatter Plot K-W Arguments
        """
        return self.main_data.plot.scatter(x=x, y=y, **kw)


class HistoryExplorer(DataExplorer):
    """
    Plotter class for data visualization of the History dataset.
    """
    @property
    def main_data(self) -> pd.DataFrame:
        return self.opt.datasets.history

    def history(self,
                column: str,
                metric: Callable[[object], object] = max,
                secondary_metric: Callable[[object], object] | None = None,
                gcol: str = 'generation',
                **kw) -> matplotlib.axes.Axes:
        """
        Scatter plot of the evolution of through multiple generations.
        Parameters
        ----------
        :column:            `str` for the Y axis column.
        :metric:            `function` (max, min, numpy.mean, custom, etc.)
        :secondary_metric:  `function` (max, min, numpy.mean, custom, etc.)
        :gcol:              Name of the 'generation' column (default: 'generation').
        Key-Word Arguments - Pandas DataFrame Scatter Plot K-W Arguments
        """
        df = self.main_data.groupby(gcol).agg(metric).reset_index()
        ax = df.plot.scatter(x=gcol, y=column, **kw)
        if secondary_metric is not None:
            df = self\
                .main_data\
                .groupby(gcol)\
                .agg(secondary_metric)\
                .reset_index()
            ax = df.plot.scatter(x=gcol, y=column, ax=ax, color='r', **kw)
        return ax


class ParetoExplorer(DataExplorer):
    """
    Plotter class for data visualization of the Simulation dataset for the Pareto Criteria.
    """
    @property
    def main_data(self) -> pd.DataFrame:
        return self.opt.datasets.simulation

    def fronts(self,
               x: str,
               y: str,
               hlfront: int | list | None = None,
               hlcolor: int | list = 'k',
               show_worst: bool = False,
               colormap: str = 'rainbow',
               frontcol: str = 'Front',
               **kw) -> matplotlib.axes.Axes:
        """
        Plot fronts in a 2D scatter plot the separates each front by color and
        highlight one specific front (if desired).

        Parameters
        ----------
        :x:             `str` for the X axis column.
        :y:             `str` for the Y axis column.
        :hlfront:       `int` or `list` for setting the highlighted front(s).
                        (default: None; if None is set, no fronts will be highlighted).
        :hlcolor:       `str` or `list` for setting the color of the highlighted front
                        (default: 'k').
        :show_worst:    `bool` set True to show the elements that didn't make the cut
                        (default: False).
        :colormap:      `str` for the desired colormap (default: 'rainbow').
        :frontcol:      Name of the 'Front' column (default: 'Front').
        Key-Word Arguments - Pandas DataFrame Scatter Plot K-W Arguments
        """
        dfplot = make_dfplot(self.main_data, show_worst)
        ax = dfplot.plot.scatter(x=x, y=y,
                                 c=frontcol,
                                 colormap=colormap,
                                 sharex=False,
                                 **kw)
        hlfront = set_hlfront(hlfront)
        if hlfront is not None:
            hlcolor = set_hlcolor(hlfront, hlcolor)
            ax = highlight_fronts(dfplot, x, y, hlfront, hlcolor, ax)
        return ax


def make_dfplot(df, show_worst, frontcol):
    if not show_worst:
        worst = df[frontcol].max()
        f = df[frontcol] != worst
        return df[f]
    else:
        return df


def set_hlfront(hlfront):
    if isinstance(hlfront, int):
        hlfront = [hlfront]
    return hlfront


def set_hlcolor(hlfront, hlcolor):
    if isinstance(hlcolor, str):
        return [hlcolor] * len(hlfront)
    elif len(hlcolor) == len(hlfront):
        return hlcolor
    else:
        s = f"""
        Number of `hlfront`s and `hlcolor`s differ.
        (hlfronts: {hlfront} | hlcolors: {hlcolor})
        """
        raise Exception(s)


def highlight_fronts(df, x, y, hlfronts, hlcolors, ax):
    fronts = df['front']
    for front, color in zip(hlfronts, hlcolors):
        ax = df[fronts == front]\
            .plot\
            .scatter(x=x, y=y, ax=ax, color=color)
    return ax



from __future__ import annotations

import numpy as np
import pandas as pd

from lamarck.genes import GeneCollection
from lamarck.population import Blueprint, Population
from lamarck.utils import ParentOverloadException


def select_parent(pop_data, n, rank_column):
    popsize = len(pop_data)
    idx = np.random.choice(popsize, n, replace=False)
    candidates = pop_data.iloc[idx]
    return candidates.sort_values(rank_column).iloc[0]


def select_parents_by_tournament(ranked_pop_data: pd.DataFrame,
                                 n_dispute: int,
                                 n_parents: int = 2,
                                 rank_column: str = 'Rank',
                                 seed: int | None = None) -> pd.DataFrame:
    """
    Select :n_parents: Parents by randomly selecting groups of :n_dispute: creatures and
    picking the highest ranked of them.

    Parameters
    ----------
    :ranked_pop_data:   A Population with a Rank column to determine the fittest creatures.
    :n_dispute:         Number of Creatures disputing for one of the Parent slots (default: 2).
    :n_parents:         Number of generated parents (default: 2).
    :rank_column:       Name of the ranking column (default: 'Rank').
    :seed:              Random Number Generator control (default: None).
    """
    np.random.seed(seed)

    # check parameters validity
    minumum_population_required = n_parents + n_dispute - 1
    total_population = len(ranked_pop_data)
    if minumum_population_required > total_population:
        raise ParentOverloadException(total_population, minumum_population_required)

    parent_index = []
    for _ in range(n_parents):
        valid_index = set(ranked_pop_data.index).difference(parent_index)
        parent = select_parent(ranked_pop_data.loc[valid_index], n_dispute, rank_column)
        parent_index.append(parent.name)
    return ranked_pop_data.loc[parent_index]


class ChildGenerator:
    """
    Generate children sexually or asexually.
    """
    @staticmethod
    def sexual(parents: pd.DataFrame,
               gene_collection: GeneCollection,
               n_children: int = 2,
               seed: int | None = None) -> pd.DataFrame:
        """
        Sexually generates offspring based on a the :parents: DataFrame.

        Parameters
        ----------
        :parents:           A Population with a Rank column to determine the fittest creatures.
        :gene_collection:   The set of `Genes` that will be crossed and mix for generating the
                            children.
        :n_children:        Number of children generated by each parent relation (default: 2).
        :seed:              Random Number Generator control (default: None).
        """
        np.random.seed(seed)
        children_data = {}
        for gene in gene_collection:
            parents_genes = tuple(parents[gene.name].tolist())
            children_genes = gene.reproduce.sexual(parents_genes, n_children)
            children_data.update({gene.name: children_genes})
        offspring = pd.DataFrame(children_data).drop_duplicates()
        offspring_size = len(offspring)
        if offspring_size < n_children:
            offspring = ChildGenerator.sexual(parents,
                                              gene_collection,
                                              n_children)
        return offspring

    @staticmethod
    def asexual(parent_creature: pd.Series,
                gene_collection: GeneCollection,
                n_children: int = 1,
                n_genes: int = 1,
                seed: int | None = None) -> pd.DataFrame:
        """
        Asexually generates offspring based on a the :parent_creature: Series.

        Parameters
        ----------
        :parent_creature:   A Pandas Series with the Creature data.
        :gene_collection:   The set of `Genes` that will be mutated for generating the children.
        :n_children:        Number of children generated by each parent relation (default: 2).
        :n_genes:           Number of Genes that will be randomly selected for mutation.
        :seed:              Random Number Generator control (default: None).
        """
        np.random.seed(seed)
        children_data = {gene: [val]*n_children for gene, val in parent_creature.iteritems()
                         if gene in gene_collection.names}
        for i_children in range(n_children):
            genes_to_mutate = np.random.choice(gene_collection, size=n_genes, replace=False)
            for gene in genes_to_mutate:
                new_gene = gene.reproduce.asexual(1)[0]
                children_data[gene.name][i_children] = new_gene
        offspring = pd.DataFrame(children_data).drop_duplicates()
        offspring_size = len(offspring)
        if offspring_size < n_children:
            offspring = ChildGenerator.asexual(parent_creature,
                                               gene_collection,
                                               n_children,
                                               n_genes)
        return offspring


child_generator = ChildGenerator()


class Populator:
    """
    Class for generating new Creaturs through Sexual and Asexual (exclusive Mutations)
    reproductions. The mutations may also occur on some of the newly generated offspring
    by random chance.

    Sexual Reproduction
        - Mix genes from multiple parents (2 or more), generating children (2 or more) with the
          new mixed genomes.

    Asexual Reproduction
        - Generate offspring by just forcing mutations on some genes from a single parent
          creature.
    """
    blueprint: Blueprint

    def __init__(self, blueprint):
        self.blueprint = blueprint

    def sexual(self,
               ranked_pop_data: pd.DataFrame,
               n_offspring: int,
               n_dispute: int = 2,
               n_parents: int = 2,
               children_per_relation: int = 2,
               rank_column: str = 'Rank',
               seed: int | None = None) -> Population:
        """
        Sexually generates :n_offspring: children based on a ranked DataFrame.

        Parameters
        ----------
        :ranked_pop_data:       A Population with a Rank column to determine the fittest
                                creatures.
        :n_offspring:           Total number of children generated by this routine (may generate
                                identical children).
        :n_dispute:             Number of Creatures disputing for one of the Parent slots
                                (default: 2).
        :n_parents:             Number of parents selected by each 'tournament' (default: 2).
        :children_per_relation: Number of children generated by each parent relation
                                (default: 2).
        :rank_column:           Name of the ranking column (default: 'Rank').
        :seed:                  Random Number Generator control (default: None).
        """
        np.random.seed(seed)
        n_relations = np.ceil(n_offspring / children_per_relation).astype(int)
        offspring_list = []
        for _ in range(n_relations):
            parents = select_parents_by_tournament(
                ranked_pop_data, n_dispute, n_parents, rank_column)
            kids = child_generator.sexual(parents,
                                          self.blueprint.genes,
                                          children_per_relation)
            offspring_list.append(kids)
        offspring_data = pd.concat(offspring_list)[self.blueprint.genes.names]\
            .reset_index(drop=True)
        return Population(offspring_data, self.blueprint.copy())

    def asexual(self,
                ranked_pop_data: pd.DataFrame,
                n_offspring: int,
                n_mutated_genes: int = 1,
                children_per_creature: int = 1,
                rank_column: str = 'Rank',
                seed: int | None = None) -> Population:
        """
        Asexually generates :n_offspring: children based on a ranked DataFrame.

        Parameters
        ----------
        :ranked_pop_data:       A Population with a Rank column to determine the fittest
                                creatures.
        :n_offspring:           Total number of children generated by this routine (may generate
                                identical children).
        :n_mutated_genes:       Number of genes that will be randomly selected and mutaded (may
                                mutate to the same value).
        :children_per_creature: Number of children generated by each creature mutation
                                (default: 1).
        :rank_column:           Name of the ranking column (default: 'Rank').
        :seed:                  Random Number Generator control (default: None).
        """
        np.random.seed(seed)
        n_mutations = np.ceil(n_offspring / children_per_creature).astype(int)
        offspring_list = []
        for _ in range(n_mutations):
            parent_creature = select_parent(ranked_pop_data, 1, rank_column)
            kids = child_generator.asexual(parent_creature,
                                           self.blueprint.genes,
                                           children_per_creature,
                                           n_mutated_genes)
            offspring_list.append(kids)
        offspring_data = pd.concat(offspring_list)[self.blueprint.genes.names]\
            .reset_index(drop=True)
        return Population(offspring_data, self.blueprint.copy())



from __future__ import annotations

from abc import ABC, abstractclassmethod, abstractstaticmethod
import itertools
import math
import numpy as np

from lamarck.genes.mixer import gene_mixer


class Gene:
    """
    Base Class for the Gene.
    """

    class GeneReproductor(ABC):
        """
        Base class for the 'reproduce' structure in the Gene Base Class.
        """
        _gene: Gene

        def asexual(self, n: int, seed: int | None = None) -> tuple:
            return tuple(self._gene.get_random(n, seed))

        @abstractstaticmethod
        def sexual(parent_genes: tuple, n_children: int = 2, seed: int | None = None) -> tuple:
            pass

    name: str
    domain: tuple
    reproduce: GeneReproductor

    @abstractclassmethod
    def get_linspace(self, n: int) -> np.ndarray:
        pass

    @abstractclassmethod
    def get_random(self, n: int, seed: int | None = None) -> np.ndarray:
        pass


class ScalarGene(Gene):
    """
    Base class for the Scalar Genes:

    Scalar Genes
    ------------
    - IntegerGene
    - FloatGene
    - CategoricalGene
    - BooleanGene
    """
    class GeneReproductor(Gene.GeneReproductor):
        _gene: ScalarGene

        def __init__(self, gene):
            self._gene = gene

        @staticmethod
        def sexual(parent_genes: tuple, n_children: int = 2, seed: int | None = None) -> tuple:
            return gene_mixer.scalar_mix(parent_genes, n_children, seed)

    name: str
    domain: tuple
    reproduce: GeneReproductor

    def __init__(self, name: str, domain: tuple):
        self.name = name
        self.domain = domain
        self.reproduce = self.GeneReproductor(self)


class IntegerGene(ScalarGene):
    """
    Integer Gene Type.

    Attributes
    ----------
    :name:      `str` Gene name.
    :domain:    `tuple` with min and max values.

    Example
    -------
    int_gene = IntegerGene(name='int_gene', domain=(1, 10))
    """
    def get_linspace(self, n: int) -> np.ndarray:
        start, end = sorted(self.domain)
        return np.linspace(start, end, n, dtype=int)

    def get_random(self, n: int, seed: int | None = None) -> np.ndarray:
        np.random.seed(seed)
        start, end = sorted(self.domain)
        return np.random.randint(start, end+1, n)


class FloatGene(ScalarGene):
    """
    Float Gene Type.

    Attributes
    ----------
    :name:      `str` Gene name.
    :domain:    `tuple` with min and max values.

    Example
    -------
    float_gene = FloatGene(name='float_gene', domain=(1., 15.5))
    """
    def get_linspace(self, n: int) -> np.ndarray:
        start, end = sorted(self.domain)
        return np.linspace(start, end, n, dtype=float)

    def get_random(self, n: int, seed: int | None = None) -> np.ndarray:
        np.random.seed(seed)
        start, end = sorted(self.domain)
        return np.random.uniform(start, end, n)


class CategoricalGene(ScalarGene):
    """
    Categorical Gene Type.

    Attributes
    ----------
    :name:      `str` Gene name.
    :domain:    `tuple` with all categories. The "categories" can consist of just any kind of
                Python object.

    Example
    -------
    cat_gene = CategoricalGene(name='cat_gene', domain=('A', 'B', 'C'))
    """
    def get_linspace(self, n: int) -> np.ndarray:
        domain_array = np.array(self.domain)
        end = len(domain_array) - 1
        index = np.linspace(0, end+1, n+1, dtype=int)[:-1]
        return domain_array[index]

    def get_random(self, n: int, seed: int | None = None) -> np.ndarray:
        np.random.seed(seed)
        return np.random.choice(self.domain, n)


class BooleanGene(ScalarGene):
    """
    Boolean Gene Type.

    Attributes
    ----------
    :name:      `str` Gene name.
    :domain:    `tuple` (False, True).

    Example
    -------
    bool_gene = BooleanGene(name='bool_gene')
    """
    def __init__(self, name: int):
        super().__init__(name=name, domain=(False, True))

    def get_linspace(self, n: int) -> np.ndarray:
        domain_array = np.array(self.domain)
        index = np.linspace(0, 2, n+1, dtype=int)[:-1]
        return domain_array[index]

    def get_random(self, n: int, seed: int | None = None) -> np.ndarray:
        np.random.seed(seed)
        return np.random.randint(0, 2, n, dtype=bool)


class VectorialGene(Gene):
    """
    Base class for the Vectorial Genes:

    Vectorial Genes
    ---------------
    - ArrayGene
    - SetGene
    """
    class GeneReproductor(Gene.GeneReproductor):
        pass

    name: str
    domain: tuple
    length: int
    reproduce: GeneReproductor

    def __init__(self, name: str, domain: tuple, length: int):
        self.name = name
        self.domain = domain
        self.length = length
        self.reproduce = self.GeneReproductor(self)

    @staticmethod
    def _convert_to_array(ls):
        arr = np.empty(len(ls), dtype=object)
        arr[:] = ls
        return arr


class ArrayGene(VectorialGene):
    """
    Array Gene Type.

    Attributes
    ----------
    name:       `str` Gene name.
    :domain:    `tuple` with all categories. The "categories" can consist of just any kind of
                Python object.
    :length:    `int` value of the array length.

    Example
    -------
    array_gene = ArrayGene(name='array_gene', domain=('i', 'j', 'k'), length=10)
    """
    class GeneReproductor(Gene.GeneReproductor):
        _gene: ArrayGene

        def __init__(self, gene):
            self._gene = gene

        @staticmethod
        def sexual(parent_genes: tuple, n_children: int = 2, seed: int | None = None) -> tuple:
            return gene_mixer.vectorial_cross(parent_genes, n_children, seed)

    def get_linspace(self, n: int) -> np.ndarray:
        vectors = list(itertools.product(self.domain, repeat=self.length))
        end = len(self.domain)**self.length - 1
        domain_array = self._convert_to_array(vectors)
        index = np.linspace(0, end, n, dtype=int)
        return domain_array[index]

    def get_random(self, n: int, seed: int | None = None) -> np.ndarray:
        np.random.seed(seed)
        a = np.array(self.domain, dtype=object)
        vectors = [tuple(np.random.choice(a, self.length, replace=True))
                   for _ in range(n)]
        return self._convert_to_array(vectors)


class SetGene(VectorialGene):
    """
    Set Gene Type.

    Attributes
    ----------
    name:       `str` Gene name.
    :domain:    `tuple` with all categories. The "categories" can consist of just any kind of
                Python object.
    :length:    `int` value of the array length. The SetGene's :length: cannot be greater than
                the length of its :domain: since the vectors are made WITHOUT replacement. It
                IS possible to create domains with repeated values though (e.g. (1, 2, 2, 3),
                ('a', 'a', 'b'), etc).

    Example
    -------
    set_gene = ArrayGene(name='set_gene', domain=('i', 'j', 'k'), length=3)
    """
    class GeneReproductor(Gene.GeneReproductor):
        _gene: ArrayGene

        def __init__(self, gene):
            self._gene = gene

        @staticmethod
        def sexual(parent_genes: tuple, n_children: int = 2, seed: int | None = None) -> tuple:
            return gene_mixer.vectorial_cross_unique(parent_genes, n_children, seed)

    def get_linspace(self, n: int) -> np.ndarray:
        vectors = list(itertools.permutations(self.domain, r=self.length))
        end = math.perm(len(self.domain), self.length) - 1
        domain_array = self._convert_to_array(vectors)
        index = np.linspace(0, end, n, dtype=int)
        return domain_array[index]

    def get_random(self, n: int, seed: int | None = None) -> np.ndarray:
        np.random.seed(seed)
        a = np.array(self.domain, dtype=object)
        vectors = [tuple(np.random.choice(a, self.length, replace=False))
                   for _ in range(n)]
        return self._convert_to_array(vectors)


genetype_dict = {
    'integer': IntegerGene,
    'float': FloatGene,
    'categorical': CategoricalGene,
    'boolean': BooleanGene,
    'array': ArrayGene,
    'set': SetGene
}


class GeneCollection:
    """
    Gene Collection.
    """
    _dict: dict

    def __init__(self, blueprint_dict):
        self._dict = {}
        for genename, geneconfig in blueprint_dict.items():
            geneclass = genetype_dict[geneconfig['type']]
            specs = geneconfig['specs']
            gene = geneclass(genename, **specs)
            self._dict.update({genename: gene})
        self.__dict__.update(self._dict)

    def __getitem__(self, key):
        if isinstance(key, int):
            key = list(self._dict)[key]
        return self._dict[key]

    def __len__(self):
        return len(self._dict)

    @property
    def names(self):
        return list(self._dict)



import hashlib
import pandas as pd


# Objective map
objective_ascending_map = {'min': True, 'max': False}


# functions
def hash_tuple(obj: tuple) -> str:
    """
    Custom hashing function.
    """
    s = str(obj).encode()
    return hashlib.sha256(s).hexdigest()


def hash_cols(df: pd.DataFrame) -> pd.Series:
    """
    Transform row data in Tuples and hash them.
    """
    cols = (df[col] for col in df.columns)
    return pd.Series(tuple(zip(*cols)), name='id').apply(hash_tuple)


def get_outputs(func):
    outputs = func.__code__.co_consts
    if isinstance(outputs[-1], tuple):
        return outputs[-1]
    else:
        return (outputs[-1],)


# Exceptions
class VectorialOverloadException(Exception):
    """
    Exception is raised when an impossible vector is declared in the gene specs.
    """

    def __init__(self, set_length, domain_lenght):
        message = (
            f"`SetGene`'s length ({set_length}) is greater than the domain's "
            f"({domain_lenght}), therefore its impossible to assign a SetGene with "
            "those specifications."
        )
        super().__init__(message)


class ParentOverloadException(Exception):
    """
    Exception is raised when an impossible number of parents is somehow solicited.
    """

    def __init__(self, total_parents, solicited_parents):
        message = (
            f"Total number of parents ({total_parents}) is lower than the number of different "
            "parents required to fulfill the specifications (in that case, a number of "
            f"{solicited_parents} is required)."
        )
        super().__init__(message)



from __future__ import annotations

import numpy as np


class GeneMixer:
    """
    Sexual reproduction class that combine parents's genes by "mixing" or "crossing" them.

    The difference between "mix" and "cross" here is that "mixing" just gets any gene from
    random parents and picks one of them, while "crossing" means the combination of two
    vectorial genes by selecting a cut point and slicing 2 vectors in two and swapping the
    latter segment from the first vector with the latter segment from the second vector.

    Methods
    -------
    - scalar_mix()
    - vectorial_mix()
    - vectorial_cross()
    - vectorial_cross_unique()
    """
    @staticmethod
    def scalar_mix(parent_genes: tuple,
                   n_children: int = 2,
                   seed: int | None = None) -> tuple:
        """
        Mix values of non-vectorial genes (may return exact same `tuple` as the :parents:
        `tuple` because it's completely random)

        Parameters
        ----------
        :parent_genes:  Parent genes as a tuple of values
        :n_children:    Number of generated children genes (default: 2)
        :seed:          Random Number Generator control (default: None)
        """
        np.random.seed(seed)
        return tuple(np.random.choice(parent_genes) for _ in range(n_children))

    @staticmethod
    def vectorial_mix(parent_genes: tuple(tuple),
                      n_children: int = 2,
                      seed: int | None = None) -> tuple(tuple):
        """
        Mix values of vectorial genes (may return exact same `tuple` as the :parents: `tuple`
        because it's completely random)

        Parameters
        ----------
        :parent_genes:  Parent genes as a tuple of values
        :n_children:    Number of generated children genes (default: 2)
        :seed:          Random Number Generator control (default: None)
        """
        def get_mix(vectors):
            m = np.array(vectors)
            idx0 = np.random.choice(a=m.shape[0], size=m.shape[1])
            idx1 = tuple(range(m.shape[1]))
            return tuple(m[idx0, idx1])

        np.random.seed(seed)
        return tuple(get_mix(parent_genes) for _ in range(n_children))

    @staticmethod
    def vectorial_cross(parent_genes: tuple(tuple),
                        n_children: int = 2,
                        seed: int | None = None) -> tuple(tuple):
        """
        Mix values of vectorial genes (may return exact same `tuple` as the :parents: `tuple`
        because it's completely random) by crossing values in no-replacement vectors.

        Parameters
        ----------
        :parent_genes:  Parent genes as a tuple of values
        :n_children:    Number of generated children genes (default: 2)
        :seed:          Random Number Generator control (default: None)
        """
        def get_random_cross_over(v1, v2):
            cutpoint = np.random.randint(1, len(v1))
            return v1[0:cutpoint] + v2[cutpoint:]

        def get_mix(vectors):
            idx = np.random.choice(a=len(vectors), size=2, replace=False)
            v1, v2 = [tuple(v) for v in np.array(vectors)[idx]]
            return get_random_cross_over(v1, v2)

        np.random.seed(seed)
        return tuple(get_mix(parent_genes) for _ in range(n_children))

    @staticmethod
    def vectorial_cross_unique(parent_genes: tuple(tuple),
                               n_children: int = 2,
                               seed: int | None = None) -> tuple(tuple):
        """
        Mix values of vectorial genes (may return exact same `tuple` as the :parents: `tuple`
        because it's completely random) by crossing values in no-replacement vectors.

        Parameters
        ----------
        :parent_genes:  Parent genes as a tuple of values
        :n_children:    Number of generated children genes (default: 2)
        :seed:          Random Number Generator control (default: None)
        """
        def get_random_cross_over(v1, v2):
            cutpoint = np.random.randint(1, len(v1))
            g1 = v1[0:cutpoint]
            rem = [v for v in v2 if v not in g1]
            remlen = len(v1) - cutpoint
            g2 = tuple(rem[0:remlen])
            return g1 + g2

        def get_mix(vectors):
            idx = np.random.choice(a=len(vectors), size=2, replace=False)
            v1, v2 = [tuple(v) for v in np.array(vectors)[idx]]
            return get_random_cross_over(v1, v2)

        np.random.seed(seed)
        return tuple(get_mix(parent_genes) for _ in range(n_children))


gene_mixer = GeneMixer()



from lamarck.population import BlueprintBuilder, Blueprint
from lamarck.optimizer import Optimizer
from lamarck.explorer import DataExplorer, HistoryExplorer, ParetoExplorer
from lamarck.version import __version__


__all__ = [
    '__version__',
    'BlueprintBuilder', 'Blueprint', 'Optimizer',
    'DataExplorer', 'HistoryExplorer', 'ParetoExplorer'
]



from __future__ import annotations
import numpy as np
import pandas as pd

from lamarck.utils import objective_ascending_map


def rank_formatter(name):
    def deco(rank_func):
        def wrapper(obj, *a, **kw):
            return rank_func(obj, *a, **kw).astype(int).rename(name)
        return wrapper
    return deco


class RankCalculator:
    """
    Fitness calculations based on the simulation results.
    """
    results: pd.DataFrame
    out: list | str

    def __init__(self, results_df: pd.DataFrame = pd.DataFrame(), out: list | str = ''):
        self.update(results_df=results_df, out=out)

    def update(self, results_df: pd.DataFrame = None, out: list | str = None) -> None:
        if results_df is not None:
            self.results = results_df.copy()
        if out is not None:
            self.out = out

    @rank_formatter('Rank')
    def single(self, objective: str) -> pd.Series:
        """
        Ranks one `output` to optimize according to a defined `objective`.
        """
        return self.results[self.out]\
            .rank(method='min', ascending=objective_ascending_map[objective])

    @rank_formatter('Rank')
    def ranked(self, objectives: list[str]) -> pd.Series:
        """
        Get the Gene Ranks based on a set of `outputs` and `objectives` in order of priority.
        """
        ranks = [
            self.results[priority].rank(method='min',
                                        ascending=objective_ascending_map[objective])
            for priority, objective in zip(self.out, objectives)]
        rank = ranks[-1]
        for r in ranks[::-1]:
            order = int(np.log10(r.max())) + 1
            factor = 10**order
            rscore = r * factor + rank
            rank = rscore.rank(method='min')
        return rank

    @rank_formatter('Rank')
    def pareto(self, objectives: list[str]) -> pd.Series:
        """
        Get the Pareto Ranks based on the `pareto fronts` and the `crowds` Series.
        """
        fronts = self.pareto_fronts(objectives)
        crowds = self.pareto_crowds(fronts)
        r1 = fronts.rank(method='dense', ascending=True)
        r2 = crowds.rank(method='dense', ascending=False)
        order1 = int(np.log10(r2.max())) + 1
        factor1 = 10**order1
        return (r1 * factor1 + r2).rank(method='min')

    def pareto_fronts(self, objectives: list[str]) -> pd.Series:
        """
        Get the Pareto Fronts.
        """
        norm_df = normalize_df_by_objective(self.results, self.out, objectives)
        dominators = get_dominators(norm_df)
        return get_fronts(dominators).rename('Front')

    def pareto_crowds(self, fronts: pd.Series) -> pd.Series:
        """
        Get the Pareto Crowds.
        """
        frontvals = sorted(fronts.unique())
        crowds = pd.Series(np.zeros(len(self.results[self.out])), index=self.results.index)
        for front in frontvals:
            f = fronts == front
            crowds[f] = get_crowd(self.results[f])
        return crowds.rename('Crowd')


def normalize_series_by_objective(series, objective):
    maxval = series.max()
    minval = series.min()
    data_range = maxval - minval
    abs_series = series - minval
    if objective == 'max':
        norm_series = abs_series/data_range
    elif objective == 'min':
        norm_series = 1 - abs_series/data_range
    return norm_series


def normalize_df_by_objective(df, outputs, objectives):
    data_dict = {
        output: normalize_series_by_objective(df[output], objective)
        for output, objective in zip(outputs, objectives)
    }
    return pd.DataFrame(data_dict, index=df.index)


def get_dominators(normalized_df):
    """
    Get the `dominators` based on the `nomalized_by_objective` df.
    """
    def dominator_mapper(row):
        diff = normalized_df - row
        f_equals = (diff == 0).all(axis=1)
        f_dominant = (diff >= 0).all(axis=1)
        return normalized_df.index[~f_equals & f_dominant]
    return normalized_df.apply(dominator_mapper, axis=1)


def get_fronts(dominators):
    """
    Get the array of `front` values base on the `dominators` array.
    """
    def isin_deco(arr):
        def isin(row):
            return row.isin(arr).all()
        return isin

    dom_arr = np.array([])
    front = 1
    fronts = pd.Series(np.zeros(len(dominators)), index=dominators.index)
    for _ in range(9):
        isin = isin_deco(dom_arr)
        f = dominators.apply(isin) & (fronts == 0)
        fronts[f] = front
        dom_arr = np.concatenate((dom_arr, f[f].index.to_numpy()))
        front += 1
    fronts[fronts == 0] = front
    return fronts.astype(int)


def get_crowd(df):
    s = pd.Series(np.zeros(len(df)), index=df.index)
    for _, cs in df.iteritems():
        infval = pd.Series([np.inf])
        si = pd\
            .concat([-infval, cs, infval])\
            .sort_values()
        sfvals = si[2:].values - si[:-2].values
        sf = pd.Series(sfvals, index=si.index[1:-1])
        s += sf
    return s



from lamarck.genes.gene import (IntegerGene, FloatGene,
                                CategoricalGene, BooleanGene,
                                ArrayGene, SetGene,
                                GeneCollection)

__all__ = [
    'IntegerGene',
    'FloatGene',
    'CategoricalGene',
    'BooleanGene',
    'ArrayGene',
    'SetGene',
    'GeneCollection'
]



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum


class ParallelMode(Enum):
    OFF = 0
    MULTITHREAD = 2
    MULTIPROCESS = 3


@dataclass
class SimulationConfig:
    """
    Simulation configurations.
    """
    max_generations: int = 20
    max_stall: int = 5
    p_selection: float = 0.5
    p_selection_weak: float = 0.
    randomize_to_fill_pop: bool = False
    n_dispute: int = 2
    n_parents: int = 2
    children_per_relation: int = 2
    p_mutation: float = 0.05
    max_mutated_genes: int = 1
    children_per_mutation: int = 1
    p_new_random: float = 0.
    parallel_processing: int = ParallelMode.OFF
    max_workers: int | None = None
    peek_champion_variables: list | None = None

    def set_parallel_processing(self, mode: int = 0) -> None:
        """
        Modes
        -----
            0: OFF
            1: MULTITHREAD
            2: MULTIPROCESS
        """
        selected_mode = {
            0: ParallelMode.OFF,
            1: ParallelMode.MULTITHREAD,
            2: ParallelMode.MULTIPROCESS
        }[mode]
        self.parallel_processing = selected_mode
